#include "artgl.h"


#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <malloc.h>

#ifdef CALLBACK
#define TESS_CALLBACK CALLBACK
#else
#define TESS_CALLBACK
#endif

typedef void(TESS_CALLBACK *tcbktype)();

typedef struct COMBINERECORD CombineRecord;

struct COMBINERECORD {
  unsigned int count;
  unsigned int capacity;
  double** ptr;
};

static void artgl_combine_push(CombineRecord* r, double* vp) {
  if (!r->capacity) {
    r->capacity = 8;
    r->ptr = (double**)malloc(r->capacity * sizeof(double*));
  } else if (r->capacity == r->count) {
    r->capacity *= 2;
    r->ptr = (double**)realloc(r->ptr, r->capacity * sizeof(double*));
  }
  r->ptr[r->count] = vp;
  r->count += 1; 
}

static void artgl_combine_clear(CombineRecord* r) {
  unsigned int i;
  for (i=0; i<r->count; ++i) {
    free(r->ptr[i]);
  }
  free(r->ptr);
}

void TESS_CALLBACK artgl_tess_combine_data(GLdouble coords[3],
                                            void* d[4],
                                            GLfloat w[4],
                                            void **out,
                                            void *userdata) {
  //printf("tess: combining (%lf, %lf, %lf)\n", coords[0], coords[1], coords[2]);
  CombineRecord* record = (CombineRecord*)userdata;
  double* newv = (double*)malloc(3*sizeof(double));
  artgl_combine_push(record, newv);
  newv[0] = coords[0];
  newv[1] = coords[1];
  newv[2] = coords[2];
  *out = newv;
}

/*
void TESS_CALLBACK artgl_tess_begin(GLenum type) {
  //printf("tess: beginning %d\n", (int)type);
  glBegin(type);
}

void TESS_CALLBACK artgl_tess_end() {
  //printf("tess: ending\n");
  glEnd();
}

void TESS_CALLBACK artgl_tess_vertex(void* data) {
  double* v = (double*)data;
  //printf("tess: got vertex (%lf, %lf)\n", v[0], v[1]);
  glVertex3dv(v);
}
*/

void TESS_CALLBACK artgl_tess_error(GLenum errno) {
  //printf("tess: error - %s\n", gluErrorString(errno));
}

static void artgl_tess_draw(ArtVpath* path, unsigned int rgba) {

  static GLUtesselator* t = 0;
  int count;
  int i;
  double* v;
  CombineRecord record;
  record.count = 0;
  record.capacity = 0;
  record.ptr = 0;

  if (!t) {
    t = gluNewTess();
    gluTessCallback(t, GLU_TESS_BEGIN, (tcbktype)glBegin);
    gluTessCallback(t, GLU_TESS_VERTEX, (tcbktype)glVertex3dv);
    gluTessCallback(t, GLU_TESS_END, (tcbktype)glEnd);
    gluTessCallback(t, GLU_TESS_COMBINE_DATA, (tcbktype)artgl_tess_combine_data);
    gluTessCallback(t, GLU_TESS_ERROR, (tcbktype)artgl_tess_error);
    gluTessNormal(t, 0, 0, 1);
  }

  for (i=0; path[i].code != ART_END; ++i);
  count = i;
  //printf("count = %d\n", count);

  v = (double*)alloca(count * sizeof(double) * 3);

  glColor4ub((rgba >> 24) & 0xff,
             (rgba >> 16) & 0xff,
             (rgba >>  8) & 0xff, 
             (rgba      ) & 0xff);

  //printf("tess: before, error string is %s\n", gluErrorString(glGetError()));
  //printf("tess: beginning polygon\n");
  gluTessBeginPolygon(t, &record);
  for (i=0; path[i].code != ART_END; ++i) {
    if (path[i].code == ART_MOVETO || path[i].code == ART_MOVETO_OPEN) {
      if (i) { 
        //printf("tess: ending contour\n");
        gluTessEndContour(t); 
      }
      //printf("tess: beginning contour\n");
      gluTessBeginContour(t);
    }
    v[3*i+0] = path[i].x;
    v[3*i+1] = path[i].y;
    v[3*i+2] = 0;
    //printf("tess: supplying vertex (%lf, %lf, %lf)\n", v[3*i+0], v[3*i+1], v[3*i+2]);
    gluTessVertex(t, v+3*i, v+3*i);
  }
  //printf("i = %d\n", count);
  //printf("tess: ending contour and polygon\n");
  gluTessEndContour(t);
  gluTessEndPolygon(t);

  //printf("tess: after, error string is %s\n", gluErrorString(glGetError()));

  //printf("record.count = %d\n", record.count);
  artgl_combine_clear(&record);

  //gluDeleteTess(t);

}

void artgl_vpath_alpha(ArtVpath* vpath,
                       unsigned int rgba) {

  artgl_tess_draw(vpath, rgba);

}

void artgl_vpath_stroke_alpha(ArtVpath* vpath,
                              unsigned int rgba,
                              ArtPathStrokeJoinType join,
                              ArtPathStrokeCapType cap,
                              double line_width) {

  ArtSVP* tmpsvp = art_svp_vpath_stroke(vpath, 
                                        join, cap,
                                        line_width, 
                                        5.0, 0.25);

  artgl_svp_alpha(tmpsvp, rgba);

  art_svp_free(tmpsvp);

}

void artgl_svp_alpha(ArtSVP* svp,
                     unsigned int rgba) {

  ArtVpath* tmpvpath = art_vpath_from_svp(svp);
  artgl_vpath_alpha(tmpvpath, rgba);
  art_free(tmpvpath);

}
