#include "svgsprite.h"
#include <assert.h>
#include <expat.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/types.h>
#include "artgl.h"
#include "xmlutil.h"
#include "mzmalloc.h"
#include <GL/GL.h>


//////////////////////////////////////////////////////////////////////

typedef struct PROPINFO {
  ArtLibrary* lib;
  Prop* prop;
} PropInfo;


void prop_tag_start(void *data, const char *el, const char **attr) {

  PropInfo* info = (PropInfo*)data;
  assert(info->lib);
  assert(!info->prop);

  if (strcmp(el, "prop")) {
    fprintf(stderr, "unrecognized tag %s in prop.\n", el);
    exit(-1);
  }

  info->prop = mznew(Prop, 1);
  
  TinySvg* svg = artlibrary_lookup_svg(info->lib, xmlutil_lookup_attr(attr, "src"));
  info->prop->svg = svg;
  assert(svg);

  const char* fstr = xmlutil_lookup_attr(attr, "flags");
  int fl;
  if (fstr && sscanf(fstr, "0x%4X", &fl)) {
    info->prop->flags = fl;
  }

  const char* valstr = xmlutil_lookup_attr(attr, "value");
  int val;
  if (valstr && sscanf(valstr, "%d", &val)) {
    info->prop->value = val;
  }

}

void prop_tag_end(void* data, const char* el) { }

Prop* prop_copy(Prop* p) {
  assert(p && p->svg);
  Prop* rval = mznew(Prop, 1);
  memcpy(rval, p, sizeof(Prop));
  return rval;
}

Prop* prop_load(ArtLibrary* l, const char* filename) {

  assert(l);
  FILE* fp = fopen(filename, "r");
  assert(fp);

  PropInfo info;
  info.lib = l;
  info.prop = 0;
  
  char buf[1024];
  unsigned int rval;

  XML_Parser p = XML_ParserCreate(NULL);
  XML_SetElementHandler(p, prop_tag_start, prop_tag_end);
  XML_SetUserData(p, &info);

  while ( (rval = fread(buf, 1, 1024, fp)) ) {
    int pv = XML_Parse(p, buf, rval, rval < 1024);
    if (!pv) {
      fprintf(stderr, "error: %s\n", XML_ErrorString(XML_GetErrorCode(p)));
      exit(-1);
    }
  }
  
  fclose(fp);

  assert(info.prop);

  strncpy(info.prop->filename, strip_filename(filename), 64);

  return info.prop;

}

void prop_free(Prop* p) {
  mzfree(p);
}

void prop_draw(Prop* p, int invert) {
  glPushMatrix();
  glTranslated(p->x, p->y, 0);
  tinysvg_draw(p->svg, 0, invert);
  glPopMatrix();
}

//////////////////////////////////////////////////////////////////////

Sprite* sprite_create(TinySvg* svg, 
                      unsigned int n_frames,
                      double* origin_coords, 
                      double* rotations) {
  Sprite* rval = mznew(Sprite, 1);

  if (n_frames) {
    assert(origin_coords && rotations);
  } else {
    assert(!origin_coords && !rotations);
  }
  rval->ref_count = 0;
  rval->svg = svg;
  rval->n_frames = n_frames;
  rval->origin_coords = origin_coords;
  rval->rotations = 0;
  return rval;
}


void sprite_free(Sprite* s) {
  assert(s);
  if (s->n_frames) {
    mzfree(s->origin_coords);
    mzfree(s->rotations);
  } else {
    assert(!s->origin_coords && !s->rotations);
  }
  mzfree(s);
}

void sprite_keyframe(Sprite* s, 
                     unsigned int f1, unsigned int f2, 
                     double blend) {
  if (f1 >= s->n_frames || f2 >= s->n_frames) {
    return;
  }
  if (blend < 0) { blend = 0; } else if (blend > 1) { blend = 1; }
}

static void sprite_cleartransform_group(TinySvgGroup* g) {
  int i;
  art_affine_identity(g->transform);
  for (i=0; i<g->n_children; ++i) {
    TinySvgNode* n = (TinySvgNode*)(g->children[i]);
    if (n->type == TINYSVG_TYPE_GROUP) {
      sprite_cleartransform_group((TinySvgGroup*)n);
    }
  }
}

void sprite_cleartransform(Sprite* s) {
  assert(s && s->svg && s->svg->root);
  sprite_cleartransform_group(s->svg->root);
}

//////////////////////////////////////////////////////////////////////

typedef struct ACTORINFO {
  ArtLibrary* library;
  Actor* actor;
  int depth;
  char stack[5][32];
} ActorInfo;


void actor_tag_actor(ActorInfo* info, const char** attr) {
  assert(info->depth == 0 && !info->actor);
  assert(!attr || !attr[0]);
  info->actor = mznew(Actor, 1);
}

void actor_tag_sprite(ActorInfo* info, const char** attr) {

  assert(info->depth == 1 && info->actor);

  // fill in
  const char* frames = xmlutil_lookup_attr(attr, "frames");
  assert(frames);
  unsigned int n_frames = strtol(frames, NULL, 10);
  assert(!n_frames);

  const char* src = xmlutil_lookup_attr(attr, "src");
  TinySvg* svg = artlibrary_lookup_svg(info->library, src);

  if (!svg) {
    fprintf(stderr,"Could not look up SVG for '%s'", src);
    exit(-1);
  }
  assert(svg);

  Sprite* s = sprite_create(svg, 0, 0, 0);
  s->ref_count = 1;

  mzrenew(info->actor->sprites, Sprite*, 
          info->actor->n_sprites, 
          info->actor->n_sprites + 1);
  
  info->actor->sprites[info->actor->n_sprites - 1] = s;

}

void actor_tag_start(void *data, const char *el, const char **attr) {
  ActorInfo* info = (ActorInfo*)data;
  if (!strcmp(el, "actor")) {
    actor_tag_actor(info, attr);
  } else if (!strcmp(el, "sprite")) {
    actor_tag_sprite(info, attr);
  } else {
    fprintf(stderr, "unrecognized tag %s in actor.\n", el);
    exit(-1);
  }
  assert(info->depth < 5);
  strcpy(info->stack[info->depth], el);
  ++(info->depth);
}

void actor_tag_end(void* data, const char* el) {
  ActorInfo* info = (ActorInfo*)data;
  --(info->depth);
}

Actor* actor_load(ArtLibrary* l, const char* filename) {

  assert(l);
  FILE* fp = fopen(filename, "r");
  assert(fp);

  char buf[1024];
  unsigned int rval;
  ActorInfo info;
  memset(&info, 0, sizeof(info));
  info.library = l;

  XML_Parser p = XML_ParserCreate(NULL);
  XML_SetElementHandler(p, actor_tag_start, actor_tag_end);
  XML_SetUserData(p, &info);

  while ( (rval = fread(buf, 1, 1024, fp)) ) {
    int pv = XML_Parse(p, buf, rval, rval < 1024);
    if (!pv) {
      fprintf(stderr, "error: %s\n", XML_ErrorString(XML_GetErrorCode(p)));
      exit(-1);
    }
  }
  
  fclose(fp);

  assert(info.actor);
  assert(info.actor->n_sprites);

  int i;
  for (i=0; i<info.actor->n_sprites; ++i) {
    TinySvg* s = info.actor->sprites[i]->svg;
    if (s->w > info.actor->w) {
      info.actor->w = s->w;
    }
    if (s->h > info.actor->h) {
      info.actor->h = s->h;
    }
  }

  strncpy(info.actor->filename, strip_filename(filename), 64);

  return info.actor;

}

Actor* actor_copy(Actor* a) {
  Actor* rval = mznew(Actor, 1);
  memcpy(rval, a, sizeof(Actor));
  rval->sprites = mznew(Sprite*, a->n_sprites);
  int i;
  for (i=0; i<a->n_sprites; ++i) {
    rval->sprites[i] = a->sprites[i];
    ++(rval->sprites[i]->ref_count);
  }
  return rval;
}


void actor_free(Actor* a) {
  unsigned int i;
  for (i=0; i<a->n_sprites; ++i) {
    Sprite* s = a->sprites[i];
    if (! --(s->ref_count)) {
      sprite_free(s);
    }
  }
  mzfree(a->sprites);
  mzfree(a);
}

void actor_draw(Actor* a, unsigned int sprite,
                unsigned int f1, unsigned int f2, 
                double blend,
                int flip, int invert) {
                

  if (sprite >= a->n_sprites) {
    sprite = 0;
  } 

  Sprite* s = a->sprites[sprite];
  double diff = a->w - s->svg->w;
  sprite_keyframe(s, f1, f2, blend);

  glPushMatrix();
  glTranslated(a->x + 0.5*diff, a->y, 0);

  tinysvg_draw(s->svg, flip, invert);

  glPopMatrix();

  sprite_cleartransform(s);

  a->lastsprite = sprite;
  a->lastf1 = f1;
  a->lastf2 = f2;
  a->lastblend = blend;
  
}

int actor_overlap_actor(Actor* a1, Actor* a2) {
  ArtDRect r1 = { a1->x, a1->y, a1->x+a1->w, a1->y+a1->h };
  ArtDRect r2 = { a2->x, a2->y, a2->x+a2->w, a2->y+a2->h };
  ArtDRect inter;
  art_drect_intersect(&inter, &r1, &r2);
  return !art_drect_empty(&inter);
}

int actor_overlap_prop(Actor* a, Prop* p) {
  ArtDRect r1 = { a->x, a->y, a->x+a->w, a->y+a->h };
  ArtDRect r2 = { p->x, p->y, p->x+p->svg->w, p->y+p->svg->h };
  ArtDRect inter;
  art_drect_intersect(&inter, &r1, &r2);
  return !art_drect_empty(&inter);
}

int prop_hoverlap_prop(Prop* a1, Prop* a2) {
  ArtDRect r1 = { a1->x, 0, a1->x+a1->svg->w, 1 };
  ArtDRect r2 = { a2->x, 0, a2->x+a2->svg->w, 1 };
  ArtDRect inter;
  art_drect_intersect(&inter, &r1, &r2);
  return !art_drect_empty(&inter);
}
 
int actor_hoverlap_actor(Actor* a1, Actor* a2) {
  ArtDRect r1 = { a1->x, 0, a1->x+a1->w, 1 };
  ArtDRect r2 = { a2->x, 0, a2->x+a2->w, 1 };
  ArtDRect inter;
  art_drect_intersect(&inter, &r1, &r2);
  return !art_drect_empty(&inter);
}

int actor_hoverlap_prop(Actor* a, Prop* p) {
  ArtDRect r1 = { a->x, 0, a->x+a->w, 1 };
  ArtDRect r2 = { p->x, 0, p->x+p->svg->w, 1 };
  ArtDRect inter;
  art_drect_intersect(&inter, &r1, &r2);
  return !art_drect_empty(&inter);
}

//////////////////////////////////////////////////////////////////////

int filter_extension(const char* str, const char* ext) {
  assert(str && ext);
  int elen = strlen(ext);
  int slen = strlen(str);
  const char* found = strstr(str, ext);
  if (!found) { return 0; }
  int pos = found - str;
  return (pos + elen) == slen;
}

FontSet* fontset_load(const char* filename) {
  FontSet* rval = mznew(FontSet, 1);
  strncpy(rval->filename, strip_filename(filename), 64);
  rval->small = TTF_OpenFont(filename, 12);
  rval->medium = TTF_OpenFont(filename, 18);
  rval->large = TTF_OpenFont(filename, 24);
  return rval;
}

void fontset_free(FontSet* f) {
  TTF_CloseFont(f->small);
  TTF_CloseFont(f->medium);
  TTF_CloseFont(f->large);
  mzfree(f);
}

ArtLibrary* artlibrary_load() {

  ArtLibrary* rval = mznew(ArtLibrary, 1);
  char buf[1024];
  
  DIR* dir = opendir("./svg");
  if (dir) {
    struct dirent* entry;
    while ( (entry = readdir(dir)) ) {
      if (filter_extension(entry->d_name, ".svg")) {
        sprintf(buf, "%s/%s", "./svg", entry->d_name);
        TinySvg* svg = tinysvg_load(buf);
        assert(svg);
        mzrenew(rval->svgs, TinySvg*, rval->n_svgs, rval->n_svgs+1);
        rval->svgs[rval->n_svgs-1] = svg;
        fprintf(stderr, "loaded %s as %s\n", buf, svg->filename);
      }
    }
    closedir(dir);
  } 

  dir = opendir("./props");
  if (dir) {
    struct dirent* entry;
    while ( (entry = readdir(dir)) ) {
      if (filter_extension(entry->d_name, ".xml")) {
        sprintf(buf, "%s/%s", "./props", entry->d_name);
        Prop* prop = prop_load(rval, buf);
        assert(prop);
        mzrenew(rval->props, Prop*, rval->n_props, rval->n_props+1);
        rval->props[rval->n_props-1] = prop;
        fprintf(stderr, "loaded %s as %s\n", buf, prop->filename);
      }
    }
    closedir(dir);
  }

  dir = opendir("./actors");
  if (dir) {
    struct dirent* entry;
    while ( (entry = readdir(dir)) ) {
      if (filter_extension(entry->d_name, ".xml")) {
        sprintf(buf, "%s/%s", "./actors", entry->d_name);
        Actor* actor = actor_load(rval, buf);
        assert(actor);
        mzrenew(rval->actors, Actor*, rval->n_actors, rval->n_actors+1);
        rval->actors[rval->n_actors-1] = actor;
        fprintf(stderr, "loaded %s as %s\n", buf, actor->filename);
      }
    }
    closedir(dir);
  }
  
  dir = opendir("./fonts");
  if (dir) {
    struct dirent* entry;
    while ( (entry = readdir(dir)) ) {
      if (filter_extension(entry->d_name, ".ttf")) {
        sprintf(buf, "%s/%s", "./fonts", entry->d_name);
        FontSet* font = fontset_load(buf);
        assert(font);
        mzrenew(rval->fonts, FontSet*, rval->n_fonts, rval->n_fonts+1);
        rval->fonts[rval->n_fonts-1] = font;
        fprintf(stderr, "loaded %s as %s\n", buf, font->filename);
      }
    }
    closedir(dir);
  }

  return rval;

}

void artlibrary_free(ArtLibrary* l) {

  int i;

  for (i=0; i<l->n_svgs; ++i) {
    tinysvg_free(l->svgs[i]);
  }
  if (l->svgs) { mzfree(l->svgs); }

  for (i=0; i<l->n_props; ++i) {
    prop_free(l->props[i]);
  }
  if (l->props) { mzfree(l->props); }

  for (i=0; i<l->n_actors; ++i) {
    actor_free(l->actors[i]);
  }
  if (l->actors) { mzfree(l->actors); }

  for (i=0; i<l->n_fonts; ++i) {
    fontset_free(l->fonts[i]);
  }
  if (l->fonts) { mzfree(l->fonts); }

  mzfree(l);

}

TinySvg* artlibrary_lookup_svg(ArtLibrary* l, const char* filename) {
  int i;
  for (i=0; i<l->n_svgs; ++i) {
    TinySvg* svg = l->svgs[i];
    if (!strcmp(svg->filename, filename)) {
      return svg;
    }
  }
  return 0;
}

Prop* artlibrary_lookup_prop(ArtLibrary* l, const char* filename) {
  int i;
  for (i=0; i<l->n_props; ++i) {
    Prop* prop = l->props[i];
    if (!strcmp(prop->filename, filename)) {
      return prop;
        
    }
  }
  return 0;
}

Actor* artlibrary_lookup_actor(ArtLibrary* l, const char* filename) {
  int i;
  for (i=0; i<l->n_actors; ++i) {
    Actor* actor = l->actors[i];
    if (!strcmp(actor->filename, filename)) {
      return actor;
    }
  }
  return 0;
}

FontSet* artlibrary_lookup_font(ArtLibrary* l, const char* filename) {
  int i;
  for (i=0; i<l->n_fonts; ++i) {
    FontSet* font = l->fonts[i];
    if (!strcmp(font->filename, filename)) {
      return font;
    }
  }
  return 0;
}
